;;	Chris Maloney
;;	filename: renservo.asm
;;  version 1.1
;;
;;	Serial input/output as well as initialization routines have been
;;	taken directly from the Renard code by Phil Short..
;;	The following is his disclaimer and I am opting for it to be mine as well.
;;
;; filename:	renard-20071209.asm
;; Copyright (c) 2006,2007 Phil Short
;;
;;  This program is provided free for you to use in any way that you wish,
;;  subject to the laws and regulations where you are using it.  Due diligence
;;  is strongly suggested before using this code.
;;
;;  The Author makes no warranty of any kind, express or implied, with regard
;;  to this program or the documentation contained in this document.  The
;;  Author shall not be liable in any event for incidental or consequential
;;  damages in connection with, or arising out of, the furnishing, performance
;;  or use of these programs.
;;
;;  --------------------------------------------------------------------------
;;
;;	This code produces approximately a .5ms - 2.5ms pulse on each servo output
;;	with 1.5ms being centered.  Each pulse is repeated each 20ms.
;;
;;	It works by dividing the 20ms pulse into 4 "phases"(each 2.5ms).
;;	Each phase handles 2 servo outputs until all outputs have been completed.

;;  --------------------------------------------------------------------------

 #define CLOCKRATE 18432000
 #define BAUDRATE 57600

#DEFINE TMR1OFFSET		.42574
	
;;  --------------------------------------------------------------------------
;;
;;	1/10/2010 - code working and tested on breadboard
;;	2/26/2010 - Remapped Servo pins for ease of designing circuit board
;;  6/13/2010 - cleaned up code, correctly servo channel order
;;  3/7/2011  - Changed Startup value from 127 to 128
;;  4/23/2011(ver 1.1) - Added assembler option to restore the origional Renard Output Mapping
;;				for use with common Renard controllers
;;	4/30/2011(ver 1.1) - Added option for Negative Output.  This will be usefull if using
;;				regular Renard controllers with SSR Replacement boards.
;;
;;  Hardware mappings
;;  PIN 13 (RA0) - servo 7 (output) 
;;  PIN 12 (RA1) - servo 6 (output)
;;  PIN 11 (RA2) - servo 5 (output)
;;  PIN 4  (RA3) - not connected(input)
;;  PIN 3  (RA4) - servo 1 (output)
;;  PIN 10 (RC0) - servo 4 (output)
;;  PIN 9  (RC1) - servo 3 (output)
;;  PIN 8  (RC2) - servo 2 (output)
;;  PIN 7  (RC3) - servo 0 (output)
;;  PIN 6  (RC4/TX) - uart_out (output)
;;  PIN 5  (RC5/RX) - uart_in (input)
;;  PIN 2  (RA5) - clock input
;;
;;  Weak pullups are used on pin RA5 (currently a spare pin).


 LIST R=DEC
 INCLUDE "p16f688.inc"

;; To change the pinouts to the origional Renard pinouts remove ;; from RENARD_PINOUT;
;; i.e. You would use this if using the Firmware in controllers other than the Renard Servo Controller.

;; #DEFINE RENARD_PINOUT;

;; For Positive output leave as is(i.e using the Renard Servo Controller
;; For Negative output remove ;; from #DEFINE OUTPUT_NEGATIVE_TRUE
;; 		use this if using a normal Renard Controller with the Servo SSR Replacement board

;; #DEFINE OUTPUT_NEGATIVE_TRUE

;; To use the internal clock of the PIC, comment out this line, remove the
;; comments on the next line, and change the clock rate from 18432000 to
;; 8000000.

 __CONFIG _FCMEN_OFF & _IESO_OFF & _BOD_OFF & _CPD_OFF & _CP_OFF & _MCLRE_OFF & _PWRTE_ON & _WDT_OFF & _EC_OSC
 ;;__CONFIG _FCMEN_OFF & _IESO_OFF & _BOD_OFF & _CPD_OFF & _CP_OFF & _MCLRE_OFF & _PWRTE_ON & _WDT_OFF & _INTOSCIO

 CBLOCK 0x020
;; variables used by the interrupt service routine (ISR)
	w_temp,status_temp

	ctr_serv0,ctr_serv1				; Phase servos
	ctr_pulse						; full pulse ctr
	servo0,servo1,servo2,servo3		; servo settings
	servo4,servo5,servo6,servo7
	phase							; current phase					

  rx_copycount
  rx_char, rx_state	
 ENDC	
	
;; derived numbers:

;; initializer for baudrate generator (based on formula from datasheet)
	
 #define BAUD_INIT (CLOCKRATE/BAUDRATE/16 - 1)

;; ---------------------------------------------------------------------------

 org 0
	nop
	goto initialize

 org 4
	;; interrupt routine
INTERRUPT

        movwf   w_temp            ; save off current W register contents
        movf    STATUS,w          ; move status register into W register
        movwf   status_temp       ; save off contents of STATUS register

		bcf		INTCON,T0IF
		movlw	.243			; reset timer 0 preset
		movwf	TMR0
		bsf		STATUS,5
		clrf	OPTION_REG
		bcf		STATUS,5
		
		clrw
		incfsz	ctr_serv0,f
		 iorlw	b'00000101'
		incfsz	ctr_serv1,f
		 iorlw	b'00011010'
		IFDEF 	OUTPUT_NEGATIVE_TRUE
		 xorlw	0xFF				; invert mask if NEGATIVE TRUE
		 iorwf	PORTA,f
		 iorwf	PORTC,f
		ELSE		
		 andwf	PORTA,f
		 andwf	PORTC,f
		ENDIF

		incfsz	ctr_pulse,f
		 goto	isr_return
		bcf		INTCON,T0IE			; turn off TMR0 interrupt
			
isr_return		
        movf    status_temp,w     ; retrieve copy of STATUS register
        movwf   STATUS            ; restore pre-isr STATUS register contents
        swapf   w_temp,f
        swapf   w_temp,w          ; restore pre-isr W register contents
        retfie

;;
;; End of Interrupt Routine, start of Foreground code.
;;
setupTMR1	MACRO				; sets a 5 ms timer
		movlw	HIGH TMR1OFFSET
		movwf	TMR1H
		movlw	LOW	 TMR1OFFSET
		movwf	TMR1L
		movlw		b'000001'	; 1:1 Prescale
		movwf	T1CON	
	endm
initialize:
;;
;; initialize some registers
;;
	; set servos to center
	movlw	.128
	movwf	servo0
	movwf	servo1
	movwf	servo2
	movwf	servo3
	movwf	servo4
	movwf	servo5
	movwf	servo6
	movwf	servo7

	clrf	phase

	clrf	rx_state
	
;; configure the I/O ports

	bcf	STATUS,RP0	; select bank 0
	movlw	7
	movwf	CMCON0		; disable the comparators

	clrf	PORTA
	clrf	PORTC

	bsf	STATUS,RP0

	movlw	0x74
	movwf	OSCCON ^ 0x80
	
	movlw	0x20		; bit 5 is the clock input bit...just in case
	movwf	TRISA ^ 0x80	; h/w ignores bit 3 of this register

	movlw	0xF0		; upper four bits are tri-stated for uart op.
	movwf	TRISC ^ 0x80		; 
	clrf	ANSEL ^ 0x80
	bcf	OPTION_REG ^ 0x80,7	; global weak-pullup enable
	movlw	0x20		; turn on weak pullup for unused input (RA5)
	movwf	WPUA ^ 0x80
	bcf	STATUS,RP0

;; configure the UART

	clrf	BAUDCTL
	movlw	BAUD_INIT;
	movwf	SPBRG
	clrf	SPBRGH
		
	movlw	0x24		; enable 8-bit transmitter in async mode
				; (x16 baud divisor)
	movwf	TXSTA
	movlw	0x90		; enable serial port with continuous 8-bit rx
	movwf	RCSTA

	IFDEF 	OUTPUT_NEGATIVE_TRUE
		movlw	0xFF		; set all outputs high(off)
		movwf	PORTA
		movwf	PORTC
	ENDIF
	
; set timer 1
	setupTMR1
	bsf		INTCON,GIE		; enable global interrupts

main:
	; test if timer 1 flag is set
	btfsc	PIR1,TMR1IF
	goto	timer1flag
	btfss	PIR1,RCIF	; skip if UART has a rx character to process
	goto	main	; no character available

;; Rx routines taken from origional renard code

	movf    RCREG,w
	movwf	rx_char		; and save a working copy
	
	movlw	6		; Rx error mask (framing error, overrun error)
	andwf	RCSTA,w		; get (masked) rx status
	btfss	STATUS,Z
	goto	rx_error	; goto error routine
	movf	rx_char,w	; retrieve the character
	addlw	0x82		; test for sync character
	btfss	STATUS,Z	; skip if sync character
	goto	rx_not_sync	; jump to code for processing non-sync
;; 
;; process sync character
;;
	bcf	rx_state,3	; exit the discarding state
	bcf	rx_state,2	; start out just echoing data
	bsf	rx_state,1
rx_echo:	
	movf	rx_char,w	; echo most recent character to transmitter
	movwf	TXREG
	goto	main

rx_error:
	bcf	RCSTA,CREN
	bsf	RCSTA,CREN
	bsf	rx_state,3	; start discarding
	goto	main

;; 
;; process non-sync character
;;
;; rx_state:
;; xxxx1xxx -> discarding data (startup, just after rx error)
;; xxxx01xx -> echoing data (count was > 0 or after end of local data)
;; xxxx001x -> expecting command/address byte (just after SYNC)
;; xxxx0001 -> process character after an ESC character
;; xxxx0000 -> process short-form data character

rx_not_sync:
	addlw	1		; test for PAD character - discard if yes
	btfsc	STATUS,Z
	goto	main
	
	btfsc	rx_state,3	; skip if not discard state
	goto	main
	btfsc	rx_state,2	; skip if non-echo state
	goto	rx_echo		; pure echo
;;
;; handle non-echo cases
;;
	btfss	rx_state,1	; skip if first character after SYNC
	goto	decode_packet_body
decode_cmd_byte:		; first character after the SYNC
	bsf	rx_state,2	; just in case packet is not for us.
	btfss	rx_char,7
	goto	rx_echo		; long form packet, just echo it.

	decf	rx_char,f	; 
	btfsc	rx_char,7	; value 0x7F after dec?  If so, it's for us
	goto	rx_echo

	clrf	rx_state	; next state is normal body
	incf	rx_char,f

	clrf	rx_copycount
	
	goto	rx_echo		; get next character (and echo the '0x80').
;;
;; done initializing packet, now handle the data
;;
decode_packet_body:	

	btfss	rx_state,0	; skip if previous char was ESC
	goto	previous_not_ESC
	movlw	0x4E
	addwf	rx_char,f	; decode the ESC charactor
	goto	after_ESC_decoding
previous_not_ESC:	
	bsf	rx_state,0	; just in case current char is ESC
	movlw	0x7F
	xorwf	rx_char,w	; check for ESC
	btfsc	STATUS,Z	; skip if not ESC
	goto	main		; go and discard the ESC
after_ESC_decoding:
	clrf	rx_state	; normal data state

	movf	rx_copycount,w
	addwf	rx_copycount,w  ; x 2
	addwf	rx_copycount,w	; x 3
	addwf	PCL,f

	IFDEF 	RENARD_PINOUT
		movf	rx_char,w		; save 'our' data
		movwf	servo1
		goto	decode_continue
		movf	rx_char,w
		movwf	servo7
		goto	decode_continue
		movf	rx_char,w
		movwf	servo6
		goto	decode_continue
		movf	rx_char,w
		movwf	servo5
		goto	decode_continue
		movf	rx_char,w
		movwf	servo4
		goto	decode_continue
		movf	rx_char,w
		movwf	servo3
		goto	decode_continue
		movf	rx_char,w
		movwf	servo2
		goto	decode_continue
		movf	rx_char,w
		movwf	servo0
	ELSE
		movf	rx_char,w		; save 'our' data
		movwf	servo0
		goto	decode_continue
		movf	rx_char,w
		movwf	servo1
		goto	decode_continue
		movf	rx_char,w
		movwf	servo2
		goto	decode_continue
		movf	rx_char,w
		movwf	servo3
		goto	decode_continue
		movf	rx_char,w
		movwf	servo4
		goto	decode_continue
		movf	rx_char,w
		movwf	servo5
		goto	decode_continue
		movf	rx_char,w
		movwf	servo6
		goto	decode_continue
		movf	rx_char,w
		movwf	servo7
	ENDIF
	
decode_continue:
	incf	rx_copycount,f	; count of bytes copied into memory
	btfsc	rx_copycount,3
	bsf	rx_state,2	; if yes, start the 'echo' rx_state
	goto	main		; no echo !
	

;*****************************************************************************
;TIMER1FLAG
timer1flag
		;; Starts Servo pulses every 20ms
		; restart timer 1
		bcf		PIR1,TMR1IF

		
		setupTMR1

		incf	phase,f						; move to next phase
		btfss	phase,2
		 goto	phaseok
		clrf	phase						; phase rolled over

phaseok
		movf	phase,w
		addwf	PCL,f
		goto	phase0
		goto	phase1
		goto	phase2

		;phase 3
		movf	servo2,w
		movwf	ctr_serv0
		movf	servo0,w
		movwf	ctr_serv1
		IFDEF 	OUTPUT_NEGATIVE_TRUE
		  movlw	b'11110011'						; turn on outputs
		  andwf PORTC,f
		ELSE
		  movlw	b'1100'
		  iorwf	PORTC,f
		ENDIF						
		goto	setuptimer0

phase0
		movf	servo7,w
		movwf	ctr_serv0
		movf	servo6,w
		movwf	ctr_serv1
		IFDEF 	OUTPUT_NEGATIVE_TRUE
		  movlw	b'11111100'
		  andwf	PORTA,f
		ELSE
		  movlw	b'11'
		  iorwf	PORTA,f						; turn on outputs
		ENDIF
		goto	setuptimer0

phase1
		movf	servo5,w
		movwf	ctr_serv0
		movf	servo1,w
		movwf	ctr_serv1
		IFDEF 	OUTPUT_NEGATIVE_TRUE
		  movlw	b'11101011'
		  andwf PORTA,f
		ELSE
		  movlw	b'10100'
		  iorwf	PORTA,f						; turn on outputs
		ENDIF
		goto	setuptimer0

phase2
		movf	servo4,w
		movwf	ctr_serv0
		movf	servo3,w
		movwf	ctr_serv1
		IFDEF 	OUTPUT_NEGATIVE_TRUE
		  movlw b'11111100'
		  andwf	PORTC,f
		ELSE
		  movlw	b'11'
		  iorwf	PORTC,f						; turn on outputs
		ENDIF
		goto	setuptimer0
	
setuptimer0		; sets a .5 ms timer
		bcf		INTCON,T0IF		; make sure interrupt flag is off
		movlw	.113
		movwf	TMR0
		banksel OPTION_REG
		movlw	b'011'			; 1:16 prescalar
		movwf	OPTION_REG		; 
		bsf		INTCON,T0IE		; enable interrupt
		banksel	ctr_pulse

		clrf	ctr_pulse		; clear pulse counter
	
		goto 	main


  END
